#
# Copyright 2019 aiohomekit team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pytest

from aiohomekit.http import HttpStatusCodes


def test_1():
    assert HttpStatusCodes[HttpStatusCodes.INTERNAL_SERVER_ERROR] == "Internal Server Error"


def test_unknown_code():
    with pytest.raises(KeyError):
        HttpStatusCodes[99]
